package s_com.udojava.evalex;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Stack;

public class Expression {
  public static final BigDecimal PI = new BigDecimal("3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679");
  
  private MathContext mc = MathContext.DECIMAL32;
  
  private String expression = null;
  
  private List<String> rpn = null;
  
  private Map<String, Operator> operators = new HashMap<>();
  
  private Map<String, Function> functions = new HashMap<>();
  
  private Map<String, BigDecimal> variables = new HashMap<>();
  
  private final char decimalSeparator = '.';
  
  private final char minusSign = '-';
  
  public Expression(String expression) {
    this.expression = expression;
    addOperator((Operator)new Object(this, "+", 20, true));
    addOperator((Operator)new Object(this, "-", 20, true));
    addOperator((Operator)new Object(this, "*", 30, true));
    addOperator((Operator)new Object(this, "/", 30, true));
    addOperator((Operator)new Object(this, "%", 30, true));
    addOperator((Operator)new Object(this, "^", 40, false));
    addOperator((Operator)new Object(this, "&&", 4, false));
    addOperator((Operator)new Object(this, "||", 2, false));
    addOperator((Operator)new Object(this, ">", 10, false));
    addOperator((Operator)new Object(this, ">=", 10, false));
    addOperator((Operator)new Object(this, "<", 10, false));
    addOperator((Operator)new Object(this, "<=", 10, false));
    addOperator((Operator)new Object(this, "=", 7, false));
    addOperator((Operator)new Object(this, "==", 7, false));
    addOperator((Operator)new Object(this, "!=", 7, false));
    addOperator((Operator)new Object(this, "<>", 7, false));
    addFunction((Function)new Object(this, "NOT", 1));
    addFunction((Function)new Object(this, "IF", 3));
    addFunction((Function)new Object(this, "RANDOM", 0));
    addFunction((Function)new Object(this, "SIN", 1));
    addFunction((Function)new Object(this, "COS", 1));
    addFunction((Function)new Object(this, "TAN", 1));
    addFunction((Function)new Object(this, "SINH", 1));
    addFunction((Function)new Object(this, "COSH", 1));
    addFunction((Function)new Object(this, "TANH", 1));
    addFunction((Function)new Object(this, "RAD", 1));
    addFunction((Function)new Object(this, "DEG", 1));
    addFunction((Function)new Object(this, "MAX", 2));
    addFunction((Function)new Object(this, "MIN", 2));
    addFunction((Function)new Object(this, "ABS", 1));
    addFunction((Function)new Object(this, "LOG", 1));
    addFunction((Function)new Object(this, "LOG10", 1));
    addFunction((Function)new Object(this, "ROUND", 2));
    addFunction((Function)new Object(this, "FLOOR", 1));
    addFunction((Function)new Object(this, "CEILING", 1));
    addFunction((Function)new Object(this, "SQRT", 1));
    this.variables.put("PI", PI);
    this.variables.put("TRUE", BigDecimal.ONE);
    this.variables.put("FALSE", BigDecimal.ZERO);
  }
  
  private boolean isNumber(String st) {
    Objects.requireNonNull(this);
    if (st.charAt(0) == '-' && st.length() == 1)
      return false; 
    for (char ch : st.toCharArray()) {
      Objects.requireNonNull(this);
      Objects.requireNonNull(this);
      if (!Character.isDigit(ch) && ch != '-' && ch != '.')
        return false; 
    } 
    return true;
  }
  
  private List<String> shuntingYard(String expression) {
    List<String> outputQueue = new ArrayList<>();
    Stack<String> stack = new Stack<>();
    Tokenizer tokenizer = new Tokenizer(this, expression);
    String lastFunction = null;
    String previousToken = null;
    while (tokenizer.hasNext()) {
      String token = tokenizer.next();
      if (isNumber(token)) {
        outputQueue.add(token);
      } else if (this.variables.containsKey(token)) {
        outputQueue.add(token);
      } else if (this.functions.containsKey(token.toUpperCase())) {
        stack.push(token);
        lastFunction = token;
      } else if (Character.isLetter(token.charAt(0))) {
        stack.push(token);
      } else if (",".equals(token)) {
        while (!stack.isEmpty() && !"(".equals(stack.peek()))
          outputQueue.add(stack.pop()); 
        if (stack.isEmpty())
          throw new ExpressionException(this, "Parse error for function '" + lastFunction + "'"); 
      } else if (this.operators.containsKey(token)) {
        Operator o1 = this.operators.get(token);
        String token2 = stack.isEmpty() ? null : stack.peek();
        while (this.operators.containsKey(token2) && ((o1.isLeftAssoc() && o1.getPrecedence() <= ((Operator)this.operators.get(token2)).getPrecedence()) || o1.getPrecedence() < ((Operator)this.operators.get(token2)).getPrecedence())) {
          outputQueue.add(stack.pop());
          token2 = stack.isEmpty() ? null : stack.peek();
        } 
        stack.push(token);
      } else if ("(".equals(token)) {
        if (previousToken != null && 
          isNumber(previousToken))
          throw new ExpressionException(this, "Missing operator at character position " + tokenizer.getPos()); 
        stack.push(token);
      } else if (")".equals(token)) {
        while (!stack.isEmpty() && !"(".equals(stack.peek()))
          outputQueue.add(stack.pop()); 
        if (stack.isEmpty())
          throw new RuntimeException("Mismatched parentheses"); 
        stack.pop();
        if (!stack.isEmpty() && this.functions.containsKey(((String)stack.peek()).toUpperCase()))
          outputQueue.add(stack.pop()); 
      } 
      previousToken = token;
    } 
    while (!stack.isEmpty()) {
      String element = stack.pop();
      if ("(".equals(element) || ")".equals(element))
        throw new RuntimeException("Mismatched parentheses"); 
      if (!this.operators.containsKey(element))
        throw new RuntimeException("Unknown operator or function: " + element); 
      outputQueue.add(element);
    } 
    return outputQueue;
  }
  
  public BigDecimal eval() {
    Stack<BigDecimal> stack = new Stack<>();
    for (String token : getRPN()) {
      if (this.operators.containsKey(token)) {
        BigDecimal v1 = stack.pop();
        BigDecimal v2 = stack.pop();
        stack.push(((Operator)this.operators.get(token)).eval(v2, v1));
        continue;
      } 
      if (this.variables.containsKey(token)) {
        stack.push(((BigDecimal)this.variables.get(token)).round(this.mc));
        continue;
      } 
      if (this.functions.containsKey(token.toUpperCase())) {
        Function f = this.functions.get(token.toUpperCase());
        ArrayList<BigDecimal> p = new ArrayList<>(f.getNumParams());
        for (int i = 0; i < f.numParams; i++)
          p.add(0, stack.pop()); 
        BigDecimal fResult = f.eval(p);
        stack.push(fResult);
        continue;
      } 
      stack.push(new BigDecimal(token, this.mc));
    } 
    return ((BigDecimal)stack.pop()).stripTrailingZeros();
  }
  
  public s_com.udojava.evalex.Expression setPrecision(int precision) {
    this.mc = new MathContext(precision);
    return this;
  }
  
  public s_com.udojava.evalex.Expression setRoundingMode(RoundingMode roundingMode) {
    this.mc = new MathContext(this.mc.getPrecision(), roundingMode);
    return this;
  }
  
  public Operator addOperator(Operator operator) {
    return this.operators.put(operator.getOper(), operator);
  }
  
  public Function addFunction(Function function) {
    return this.functions.put(function.getName(), function);
  }
  
  public s_com.udojava.evalex.Expression setVariable(String variable, BigDecimal value) {
    this.variables.put(variable, value);
    return this;
  }
  
  public s_com.udojava.evalex.Expression setVariable(String variable, String value) {
    if (isNumber(value)) {
      this.variables.put(variable, new BigDecimal(value));
    } else {
      this.expression = this.expression.replaceAll("\\b" + variable + "\\b", "(" + value + ")");
      this.rpn = null;
    } 
    return this;
  }
  
  public s_com.udojava.evalex.Expression with(String variable, BigDecimal value) {
    return setVariable(variable, value);
  }
  
  public s_com.udojava.evalex.Expression and(String variable, String value) {
    return setVariable(variable, value);
  }
  
  public s_com.udojava.evalex.Expression and(String variable, BigDecimal value) {
    return setVariable(variable, value);
  }
  
  public s_com.udojava.evalex.Expression with(String variable, String value) {
    return setVariable(variable, value);
  }
  
  public Iterator<String> getExpressionTokenizer() {
    return (Iterator<String>)new Tokenizer(this, this.expression);
  }
  
  private List<String> getRPN() {
    if (this.rpn == null)
      this.rpn = shuntingYard(this.expression); 
    return this.rpn;
  }
  
  public String toRPN() {
    String result = new String();
    for (String st : getRPN()) {
      result = result.isEmpty() ? result : (result + " ");
      result = result + result;
    } 
    return result;
  }
}
